import yfinance as yf
import pandas as pd
import numpy as np
from statsmodels.tsa.holtwinters import ExponentialSmoothing
from sklearn.metrics import mean_absolute_error, mean_squared_error
from joblib import Parallel, delayed  # Parallel processing

# === CONFIGURATION ===
# DEFINE YOUR NUMBERS HERE (INCLUSIVE)
START_NUM = 1000
END_NUM = 1999
df = pd.read_pickle("Clean Stock_prices_Adj Close_Dropped.pyl")
df_date = df.loc[:, ["Date"]]
df_helpme = df.iloc[:, START_NUM: END_NUM]
df_select = pd.concat([df_date, df_helpme], axis=1)
ticker_symbols = df_helpme.columns.values.tolist()

# === Function to Process a Single Ticker ===
def process_ticker(ticker_symbol):
    print(f"\nProcessing: {ticker_symbol}")

    # === Step 1: Download stock data (2010–2019) ===
    try:
        # data = yf.download(ticker_symbol, start="2010-01-01", end="2020-01-01", auto_adjust=True, progress=False)
        data = df_select.loc[:, ["Date", ticker_symbol]]
        if data.empty:
            print(f"Skipping {ticker_symbol} (No data)")
            return None
    except Exception as e:
        print(f"Error fetching data for {ticker_symbol}: {e}")
        return None

    data = data.dropna()
    data.columns = ['ds', 'y']
    data['ds'] = pd.to_datetime(data['ds'])

    # === Step 2: Split into train and test ===
    train = data[data['ds'] < '2016-01-01'].copy()
    test = data[(data['ds'] >= '2016-01-01') & (data['ds'] < '2018-01-01')].copy()
    train.set_index('ds', inplace=True)
    test.set_index('ds', inplace=True)

    if len(train) < 2 or len(test) < 2:  # Ensure sufficient data points
        print(f"Skipping {ticker_symbol} (Insufficient data)")
        return None

    # === Step 3: Fit Holt-Winters model ===
    try:
        model = ExponentialSmoothing(
            train['y'], trend='mul', seasonal='mul', seasonal_periods=687
        ).fit()
    except Exception as e:
        print(f"Error training model for {ticker_symbol}: {e}")
        return None

    # === Step 4: Forecast ===
    forecast = model.forecast(len(test))
    forecast = pd.Series(forecast.values, index=test.index)

    # === Step 5: Evaluation Metrics ===
    mae = mean_absolute_error(test['y'], forecast)
    rmse = np.sqrt(mean_squared_error(test['y'], forecast))
    mape = np.mean(np.abs((test['y'] - forecast) / test['y'])) * 100
    final_actual = test['y'].iloc[-1]
    final_pred = forecast.iloc[-1]
    final_diff = abs(final_actual - final_pred)

    # === Step 6: Volatility Calculation ===
    daily_returns = data.set_index('ds')['y'].pct_change().dropna()
    volatility = daily_returns.std() * np.sqrt(252)  # Annualized volatility

    # === Console Output ===
    print(f"\n--- {ticker_symbol} Evaluation Metrics ---")
    print(f"MAE  : {mae:.2f}")
    print(f"RMSE : {rmse:.2f}")
    print(f"MAPE : {mape:.2f}%")
    print(f"Actual Closing Price  : {final_actual:.2f}")
    print(f"Predicted Closing Price: {final_pred:.2f}")
    print(f"Difference             : {final_diff:.2f}")
    print(f"Annualized Volatility  : {volatility:.2%}")

    # === Store Results ===
    return {
        "Ticker": ticker_symbol,
        "MAE": round(mae, 2),
        "RMSE": round(rmse, 2),
        "MAPE (%)": round(mape, 2),
        "Actual (End 2017)": round(final_actual, 2),
        "Predicted (End 2017)": round(final_pred, 2),
        "Difference": round(final_diff, 2),
        "Annualized Volatility (%)": round(volatility * 100, 2)
    }


# === Run in Parallel (Using 8 CPU Cores) ===
num_cores = 8  # Adjust based on your CPU capabilities
results = Parallel(n_jobs=num_cores)(delayed(process_ticker)(ticker) for ticker in ticker_symbols)
# results = [process_ticker(ticker) for ticker in ticker_symbols]

# === Remove None values and Convert to DataFrame ===
results = [r for r in results if r is not None]
results_df = pd.DataFrame(results)

# === Export to CSV ===
results_df.to_excel("holtwinters_results_parallel_new.xlsx", index=False)
print("\nResults saved to holtwinters_results_parallel_new.xlsx")